//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// common.c - common functions and definitions
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <conio.h>
#include <setjmp.h>

#include "locotool.h"
#include "common.h"

// error handler proc to call on errors
LocoTool_err_proc_t err_die;
LocoTool_err_proc err_handler = err_die;

// setjmp/longjmp buffer; if set an error will return using longjmp
jmp_buf *err_return = NULL;

// to debug the heap functions
int heapdebug = 0;

__declspec(dllexport) void LocoTool_set_err_handler(LocoTool_err_proc handler)
{
  err_handler = handler;
}

__declspec(dllexport) const char *LocoTool_get_version()
{
  return "LocoTool V0.06 - Copyright (C) 2004 by Josef Drexler";
}

void err_die(char *message)
{
  fflush(stdout);
  fflush(stderr);

  fputs(message, stderr);

  fputs("\nPress any key to abort.", stderr);

  getch();
  fputs("\n", stderr);

  exit(1);
}

__declspec(dllexport) void LocoTool_die(const char s[], ...)
{
  char message[1024];

  va_list args;

  va_start(args, s);
  vsnprintf(message, sizeof(message), s, args);
  va_end(args);

  err_handler(message);

  if (err_return) {
	longjmp(*err_return, 1);
  }

  exit(1);
}

//
// Heap functions
//
// Unlike regular malloc et.al., they die in an out-of-memory condition,
// and free and relloc also check that the pointer is a valid memory handle
//
// They do this by maintaining a 16-byte structure at the beginning of each
// memory block, and only returning a pointer past this block
//

void *my_alloc(size_t len, const char *name)
{
  void *p;

  if (heapdebug) fprintf(stderr, "my_alloc(%d, %s)", len, name);

  p = malloc(len+16);
  if (!p) die("Not enough memory for %s, wanted %d", name, len);
  memset(p, 0, len+16);

  if (heapdebug) fprintf(stderr, " = %p\n", p);

  *(uint32_t*)p = 0x2343211;
  strncpy(((char*)p)+4, name, 7);

  return ((char*)p)+16;
}

void *my_realloc(void *p, size_t len, size_t oldlen, const char *name)
{
  if (*(uint32_t*)(((char*)p)-16) != 0x2343211)
	die("Trying to realloc invalid pointer!");

  if (heapdebug) fprintf(stderr, "my_realloc(%p, %d, %d, %s) [%s]", ((char*)p)-16, len, oldlen, name, ((char*)p)-12);

  p = realloc(((char*)p)-16, len+16);
  if (!p) die("Not enough memory for %s, wanted %d", name, len);

  if (oldlen<len) memset(((char*)p)+oldlen+16, 0, len-oldlen);

  if (heapdebug) fprintf(stderr, " = %p\n", p);

  *(uint32_t*)p = 0x2343211;
  strncpy(((char*)p)+4, name, 7);

  return ((char*)p)+16;
}

char *my_strdup(const char *s, const char *name)
{
  char *c = my_alloc(strlen(s)+1,"name");
  strcpy(c, s);
  return c;
}

// this is normally used via the my_free macro which also zeroes p afterwards
void *_my_free(void *p)
{
  if (p) {
	if (*(uint32_t*)(((char*)p)-16) != 0x2343211)
		die("Trying to free invalid pointer!");
	if (heapdebug) fprintf(stderr, "my_free(%p) [%s]\n", ((char*)p)-16, ((char*)p)-12);

	free(((char*)p)-16);
  }
  return NULL;
}
