#ifndef _LOCODEF_H
#define _LOCODEF_H

//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// locodef.h - data structures and variables used across the encoding/decoding
//	       functions
//

#include <stdio.h>


// Global variables

extern int embedpng;	// embed PNG data in the XML file
extern int allinone;	// put all sprites into a single PNG file
extern char *filebase;	// .dat/.xml file name with the extension removed
extern char *pngbase;	// same, but as directory


// sprite flag definition; used by various handlers

extern const char *spriteflags[];


// enum of all supported XML elements
//
// el_document is pseudo-element of the notional document that contains
// the xml data of the file
//
// NOTE: Any changes here must be synchronized with the elemlist[] array in xml.c
//
enum en_elemid {
	el_bit, el_bitmask, el_cargo, el_cargotype, el_chunk,
	el_description, el_object, el_pngfile, el_sprite,
	el_structure, el_unknown, el_useobject, el_variable,

	el_document,
};
typedef enum en_elemid elemid;


// enum of all supported XML attributes
//
// the first set have a number as argument, the second set after at_STRINGVALUE
// have a string as argument
//
// NOTE: Any changes here must be synchronized with the attname[] array in xml.c
//
enum en_attrid {
	at_capacity, at_class, at_compression, at_id,
	at_language, at_num, at_size, at_subclass, at_xofs, at_yofs,

	at_STRINGVALUE,

	at_name, at_desc,

	at_LASTVALUE,
};
typedef enum en_attrid attrid;

extern const char *attname[at_LASTVALUE];


// enum of all support <useobject> types
//
// NOTE: Any changes here must be synchronized with the useobjnames[] array in xml.c
//
enum en_useobjid {
	ob_tracktype, ob_trackmod, ob_visualeffect, ob_wakeeffect,
	ob_rackrail, ob_compatible, ob_startsnd, ob_soundeffect,
	ob_signal, ob_tunnel, ob_bridge, ob_station,

	ob_LASTVALUE,
};
typedef enum en_useobjid useobjid;

extern const char *useobjnames[ob_LASTVALUE];


//
// definition of an object structure variable
//
struct s_varinf {
	int ofs;		// offset counted from the beginning of the structure
	int size;		// size in bytes, negative for signed variables
	int num;		// number of entries (for arrays)
	const char *name;	// name, if known; if empty (*not* NULL!) will use generic field_## name
	struct s_varinf *structvars;	// if not NULL, sub-structure definition
	const char **flags;		// if not NULL, bit field definition (pointer to list of *char bit names)
};
typedef struct s_varinf varinf;

//
// definition of a sprite
//
struct s_spriteinf {
	uint32_t width,height,flags, spritesize;
	int32_t xofs,yofs;
};
typedef struct s_spriteinf spriteinf;

//
// definition of an external object dependence
//
struct s_useobjinf {
	int class, num;
	useobjid type;
	uint8_t obj[16];	// the binary code in the .dat file
};
typedef struct s_useobjinf useobjinf;

//
// attribute value; either a number or string depending on whether it
// is defined before or after at_STRINGVAL in en_attrid
//
union u_attval {
	uint32_t u32;
	char *str;
};
typedef union u_attval attval;

//
// description of an element; can contain either a list of bit names or
// a sub-structure, depending on the element
//
union u_elemdesc {
	const char **flags;
	varinf *vars;
};
typedef union u_elemdesc elemdesc;

//
// information about an XML element
//
struct s_eleminf {
	elemid id;			// element ID
	attval atts[at_LASTVALUE];	// attribute values
	elemdesc desc;			// description of contents
	void *data;			// pointer to corresponding data in object structure
};
typedef struct s_eleminf eleminf;

//
// definition about an array holding XML information
// these arrays can be automatically made longer using the checklen() function
//
struct s_arrayinf {
		// size: number of bytes allocated to the array (may grow)
		// ent:  number of entries defined (may grow or become smaller)
		// entsize: number of bytes per entry (set at initialization)
	uint32_t size, ent, entsize;

		// pointer to array of data entries
		// e.g. for useobj_i this points to useobj
		// (set at initialization)
	char ***data;
};
typedef struct s_arrayinf arrayinf;

//
// all information collected from the XML file
//
struct s_xmlinf {

	//
	// META DATA (only used while parsing, doesn't contain object data)
	//

		// does the current element need character data?
	int chardata;

		// current object character data, if chardata is non-zero
		// (i.e. the stuff betyween <element> and </element>)
	uint32_t curvaluelen, curvaluesize;
	char *curvalue;

		// current element stack
	arrayinf elems_i;
	eleminf **elems;

	//
	// OBJECT DATA
	//

		// object class, chunk compression
	int class, compression;

		// object subclass (no idea what this means)
	uint32_t subclass;

		// object name, space padded to 8 bytes
	uint8_t name[8];

		// object data structure; size is object class specific
	arrayinf objdata_i;
	uint8_t *objdata;

		// language/description data
	arrayinf descdata_i;
	uint8_t *descdata;

		// external object dependencies
	arrayinf useobj_i;
	useobjinf **useobj;

		// capacity data for both cargo bays
	uint32_t capsize[2], caplen[2];		// data size and used length
	uint8_t *caps[2];			// capacity data

		// list of sprite information (width/height/etc)
	arrayinf sprites_i;
	spriteinf **sprites;

		// sprite data
	uint32_t cursprite;
	arrayinf spritedata_i;
	uint8_t **spritedata;
};
typedef struct s_xmlinf xmlinf;

//
// definition of the handlers to dump to xml or parse xml into the object data
//
typedef uint32_t dumpproc(FILE*, uint8_t*);
typedef void parseproc(xmlinf*);

//
// object class-specific definitions and handlers
//
struct s_objclass {
	varinf *vars;
	uint32_t size;
	dumpproc *dump;
	parseproc *parse;
};
typedef struct s_objclass objclass;

//
// array object class-specific definitions and handlers for each class
//
#define NUMOBJCLASSES 0x22
extern objclass objclasses[NUMOBJCLASSES];


#endif // ndef _LOCODEF_H
