//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// objects.c - object class-specific handlers and data structure definitions
//

#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "dump.h"
#include "locodef.h"
#include "xml.h"

//
// Two main contents of this file:
// - the object class-specific data structures
//   (in varinf arrays plsu char* arrays for the bit fields)
// - the object class-specific decode and encode handlers
//
// Each varinf entry is either a simple variable, a sub-structure, a
// bit field, or an array of those.  All must be in sequence without
// holes.
//
// The handlers really only decide on additional information to use, the
// main object data is handled through its varinf structure, but the order
// of external object definitions, descriptions and other items is also
// class specific and thus needs its own handler.
//
//
// All object classes are then put into the objclasses[] array at the bottom
// of this file.
//
// Since everything is about the same for all classes, only the vehicle
// class has useful comments, so look there for guidance.
//

// ***********************
//      General items
// ***********************

	// the known bits in the flags of each sprite
const char *spriteflags[] = {
	"hasdata", "", "chunked",
	NULL
};

// ***********************
//  SIMPLE CLASS HANDLER
// ***********************
// Class 04:   CLIFFS
// Class 0D:   TUNNELS
// Class 19:   SNOW
// ***********************

varinf simplevars[] = {
	{ 0x00, 1, 6, "" },
	{ -1 },
};

uint32_t dumpsimple(FILE *xml, uint8_t *data)
{
  uint32_t dumped;

  dumped = dumpvars(xml, data, simplevars);
  dumped += dumplang(xml, data+dumped);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsesimple(xmlinf *inf)
{
  addval(inf, inf->descdata, inf->descdata_i.ent);
  addsprites(inf);
}


// ***********************
// Class 0E:   BRIDGES
// ***********************

varinf bridgevars[] = {
	{ 0x00, 1, 2, "" },
        { 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "" },
	{ 0x05, 1, 1, "" },
	{ 0x06, 1, 1, "" },
	{ 0x07, 1, 1, "" },
	{ 0x08, 1, 1, "spanlength" },
	{ 0x09, 1, 1, "" },
	{ 0x0A, 2, 1, "maxspeed" },
	{ 0x0C, 1, 1, "maxheight" },
	{ 0x0D, 1, 1, "costind" },
	{ 0x0E,-2, 1, "basecostfact" },
	{ 0x10,-2, 1, "heightcostfact" },
	{ 0x12,-2, 1, "salecostfact" },
	{ 0x14, 2, 1, "disabledtrackcfg" },
	{ 0x16, 1,22, "" },
	{ -1 },
};

uint32_t dumpbridge(FILE *xml, uint8_t *data)
{
  int i;
  uint32_t dumped;

  dumped = dumpvars(xml, data, bridgevars);
  dumped += dumplang(xml, data+dumped);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsebridge(xmlinf *inf)
{
  int i;

  addval(inf, inf->descdata, inf->descdata_i.ent);
  addsprites(inf);
}


// ***********************
// Class 11:   TRACKS
// ***********************

varinf trackvars[] = {
	{ 0x00, 1, 7, "" },
	{ 0x07, 1, 1, "numcompat" },
	{ 0x08, 1, 1, "nummods" },
	{ 0x09, 1, 1, "numsignals" },
	{ 0x0A, 1, 10, "" },
	{ 0x14,-2, 1, "buildcostfact" },
	{ 0x16,-2, 1, "sellcostfact" },
	{ 0x18,-2, 1, "tunnelcostfact" },
	{ 0x1A, 1, 1, "costind" },
	{ 0x1B, 1, 1, "" },
	{ 0x1C, 2, 1, "curvespeed" },
	{ 0x1E, 1, 6, "" },
	{ 0x24, 1, 1, "numbridges" },
	{ 0x25, 1, 7, "" },
	{ 0x2C, 1, 1, "numstations" },
	{ 0x2D, 1, 7, "" },
	{ 0x34, 1, 1, "displayoffset" },
	{ 0x35, 1, 1, "" },
	{ -1 },
};

uint32_t dumptrack(FILE *xml, uint8_t *data)
{
  int i;
  uint32_t dumped;

  dumped = dumpvars(xml, data, trackvars);
  dumped += dumplang(xml, data+dumped);
  for (i=0; i<data[7]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[7], ob_compatible, -1);
  for (i=0; i<data[8]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[8], ob_trackmod, 0x10);
  for (i=0; i<data[9]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[9], ob_signal, 0x0A);
  dumped += dumpuseobj(xml, data+dumped, i, 0, ob_tunnel, 0x0D);
  for (i=0; i<data[0x24]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x24], ob_bridge, 0x0E);
  for (i=0; i<data[0x2C]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x2C], ob_station, 0x0F);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsetrack(xmlinf *inf)
{
  int i;

  addval(inf, inf->descdata, inf->descdata_i.ent);
  for (i=0; i<inf->objdata[7]; i++)
	addobj(inf, i, ob_compatible, -1);
  for (i=0; i<inf->objdata[8]; i++)
	addobj(inf, i, ob_trackmod, 0x10);
  for (i=0; i<inf->objdata[9]; i++)
	addobj(inf, i, ob_signal, 0x0A);
  addobj(inf, 0, ob_tunnel, 0x0D);
  for (i=0; i<inf->objdata[0x24]; i++)
	addobj(inf, i, ob_bridge, 0x0E);
  for (i=0; i<inf->objdata[0x2C]; i++)
	addobj(inf, i, ob_station, 0x0F);
  addsprites(inf);
}


// ***********************
// Class 17:   VEHICLES
// ***********************


	// the known bits in the vehicle sprite structure flags
const char *vehspriteflags[] = {
	"hassprites", "", "", "", "",
	"reversed",
	NULL
};

	// the vehicle sprite definition structure
varinf vehsprites[] = {
	{ 0x00, 1, 1, "numdir" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "" },
	{ 0x05, 1, 1, "" },
	{ 0x06, 1, 1, "bogiepos" },
	{ 0x07, 1, 1, "flags", NULL, vehspriteflags },
	{ 0x08, 1, 1, "" },
	{ 0x09, 1, 1, "" },
	{ 0x0A, 1, 1, "" },
	{ 0x0B, 1, 1, "" },
	{ 0x0C, 1, 1, "" },
	{ 0x0D, 1, 1, "" },
	{ 0x0E, 1, 1, "spritenum" },
	{ 0x0F, 1, 3, "" },
	{ 0x12, 4, 1, "" },
	{ 0x16, 4, 1, "" },
	{ 0x1A, 4, 1, "" },
	{ -1 },
};

	// a vehicle structure whose purpose is mostly unknown
varinf vehunk[] = {
	{ 0x00, 1, 1, "length" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "spriteind" },
	{ 0x05, 1, 1, "" },
	{ -1 },
};

	// the vehicle flags
const char *vehflags[] = {
	"", "", "", "",
	"", "", "rackrail", "",
	"", "anytrack", "", "",
	"dualhead", "", "refittable", "noannounce",
	NULL
};

	// the main vehicle data structure
varinf vehvars[] = {
	{ 0x000, 1, 2, "" },
	{ 0x002, 1, 1, "class" },
	{ 0x003, 1, 1, "" },
	{ 0x004, 1, 1, "" },
	{ 0x005, 1, 1, "" },
	{ 0x006, 1, 1, "nummods" },
	{ 0x007, 1, 1, "costind" },
	{ 0x008,-2, 1, "costfact" },	// size -2 means signed word
	{ 0x00A, 1, 1, "reliability" },
	{ 0x00B, 1, 1, "runcostind" },
	{ 0x00C,-2, 1, "runcostfact" },
	{ 0x00E, 1, 1, "" },
	{ 0x00F, 1, 1, "numcompat" },
	{ 0x010, 1,20, "" },
	{ 0x024, 6, 4, "", vehunk },
	{ 0x03C,30, 4, "sprites", vehsprites },
	{ 0x0B4, 1,36, "" },
	{ 0x0D8, 2, 1, "power" },
	{ 0x0DA, 2, 1, "speed" },
	{ 0x0DC, 2, 1, "rackspeed" },
	{ 0x0DE, 2, 1, "weight" },
	{ 0x0E0, 2, 1, "flags", NULL, vehflags },
	{ 0x0E2, 1,44, "" },
	{ 0x10E, 1, 1, "" },
	{ 0x10F, 1, 1, "visfxtype" },
	{ 0x110, 1, 1, "" },
	{ 0x111, 1, 1, "" },
	{ 0x112, 1, 1, "wakefxtype" },
	{ 0x113, 1, 1, "" },
	{ 0x114, 2, 1, "designed" },
	{ 0x116, 2, 1, "obsolete" },
	{ 0x118, 1, 1, "" },
	{ 0x119, 1, 1, "startsndtype" },
	{ 0x11A, 1,64, "" },
	{ 0x15A, 1, 1, "numsnd" },
	{ 0x15B, 1, 3, "" },
	{ -1 },
};

	// dump the vehicle data to the xml file
uint32_t dumpveh(FILE *xml, uint8_t *data)
{
  int i;
  uint32_t dumped;

	// start with the vehicle data structure
  dumped = dumpvars(xml, data, vehvars);

	// then the descriptions in various languages
  dumped += dumplang(xml, data+dumped);

	// rail and road vehicles may be bound to specific track types
  if ((data[2] < 2) && !(getvalue(data+0xE0, 2) & 0x200))
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_tracktype, 0x11 + data[2]*3);

	// and they may require certain track modifications
  for (i=0; i<data[6]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[6], ob_trackmod, 0x10 + data[2]*3);

	// plus they have up to two cargo bays
  for (i=0; i<2; i++)
	dumped += dumpcap(xml, data+dumped, i, 2);

	// as well as a visual effect (e.g. steam/smoke/sparks)
  if (data[0x10F])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_visualeffect, 0x03);

	// and for boats, a wake
  if (data[0x112])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_wakeeffect, 0x03);

	// trains may use a rack rail modification
  if ((data[2] < 2) && (data[0xE0]&0x40))
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_rackrail, 0x10);

	// vehicles with which this one is compatible
	// (i.e. they can be in the same train)
  for (i=0; i<data[15]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[15], ob_compatible, 0x17);

	// start sound effect
  if (data[0x119])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_startsnd, 0x01);

	// other sound effects
  for (i=0; i<(data[0x15A] & 0x7f); i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x15A] & 0x7f, ob_soundeffect, 0x01);

	// and finally the sprites
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

	// assemble the vehicle data back out of the parsed xml info
void parseveh(xmlinf *inf)
{
  int i;
  uint8_t zero;

  zero = 0;

  // the vehicle structure is already assembled in the inf->objdata array
  // we only need to assemble the rest of the items

  addval(inf, inf->descdata, inf->descdata_i.ent);
  if ((inf->objdata[2] < 2) && !(getvalue(inf->objdata+0xE0, 2) & 0x200))
	addobj(inf, 0, ob_tracktype, 0x11 + inf->objdata[2]*3);
  for (i=0; i<inf->objdata[6]; i++)
	addobj(inf, i, ob_trackmod, 0x10 + inf->objdata[2]*3);
  for (i=0; i<2; i++)
	if (inf->capsize[i])
		addval(inf, inf->caps[i], inf->caplen[i]);
	else
		addval(inf, &zero, 1);
  if (inf->objdata[0x10F])
	addobj(inf, 0, ob_visualeffect, 0x03);
  if (inf->objdata[0x112])
	addobj(inf, 0, ob_wakeeffect, 0x03);
  if ((inf->objdata[2] < 2) && (inf->objdata[0xE0]&0x40))
	addobj(inf, 0, ob_rackrail, 0x10);
  for (i=0; i<inf->objdata[15]; i++)
	addobj(inf, i, ob_compatible, 0x17);
  if (inf->objdata[0x119])
	addobj(inf, 0, ob_startsnd, 0x01);
  for (i=0; i<(inf->objdata[0x15A] & 0x7f); i++)
	addobj(inf, i, ob_soundeffect, 0x01);
  addsprites(inf);
}

// ***********************
//  END OF CLASS HANDLERS
// ***********************

//
// Array defining object class-specific data structures and handlers
//
// Entries are object structure description, its size, dump handler, parse handler
//
objclass objclasses[NUMOBJCLASSES] = {
	{ NULL, -1, NULL, NULL },			// 00 Interfaces
	{ NULL, -1, NULL, NULL },			// 01 Sound effects
	{ NULL, -1, NULL, NULL },			// 02 Currencies
	{ NULL, -1, NULL, NULL },			// 03 Exhaust effects
	{ simplevars, 6, dumpsimple, parsesimple},	// 04 Cliff faces
	{ NULL, -1, NULL, NULL },			// 05 Water
	{ NULL, -1, NULL, NULL },			// 06 Ground
	{ NULL, -1, NULL, NULL },			// 07 Town names
	{ NULL, -1, NULL, NULL },			// 08 Cargos
	{ NULL, -1, NULL, NULL },			// 09 Fences
	{ NULL, -1, NULL, NULL },			// 0A Signals
	{ NULL, -1, NULL, NULL },			// 0B Crossings
	{ NULL, -1, NULL, NULL },			// 0C Street lights
	{ simplevars, 6, dumpsimple, parsesimple},	// 0D Tunnels
	{ bridgevars, 0x2c, dumpbridge, parsebridge },	// 0E Bridges
	{ NULL, -1, NULL, NULL },			// 0F Train stations
	{ NULL, -1, NULL, NULL },			// 10 Track modifications
	{ trackvars, 0x36, dumptrack, parsetrack },	// 11 Tracks
	{ NULL, -1, NULL, NULL },			// 12 Road stations
	{ NULL, -1, NULL, NULL },			// 13 Road modifications
	{ NULL, -1, NULL, NULL },			// 14 Roads
	{ NULL, -1, NULL, NULL },			// 15 Airports
	{ NULL, -1, NULL, NULL },			// 16 Docks
	{ vehvars, 0x15e, dumpveh, parseveh },		// 17 Vehicles
	{ NULL, -1, NULL, NULL },			// 18 Trees
	{ simplevars, 6, dumpsimple, parsesimple},	// 19 Snow
	{ NULL, -1, NULL, NULL },			// 1A Climates
	{ NULL, -1, NULL, NULL },			// 1B Shapes
	{ NULL, -1, NULL, NULL },			// 1C Buildings
	{ NULL, -1, NULL, NULL },			// 1D Scaffolding
	{ NULL, -1, NULL, NULL },			// 1E Industries
	{ NULL, -1, NULL, NULL },			// 1F Landscapes
	{ NULL, -1, NULL, NULL },			// 20 Companies
	{ NULL, -1, NULL, NULL },			// 21 Textures
};

